#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "far_types.h"
#include "far.h"
#include "far_err.h"
#include "streams.h"
#include "md5.h"
#include "fileops.h"

ferr_t far_write(buf_t *dest, void *src, uint64_t size)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dest && src)
	{
		if(size)
		{
			switch(far_buftype(dest))
			{
				case _FAR_STREAM_BUF:
					if(memcpy(far_bufptr(dest), src,
						size))
					{
						far_adv_index(dest, size);
					}
					else
					{
						_far_set_errcode(result,
								errno);
					}
					break;

				case _FAR_STREAM_FD:
				case _FAR_STREAM_SD:
					result = far_fwrite(far_filedesc(dest),
								src, size);
					break;

				default:
					_far_set_errcode(result,
							_FAR_ERR_BADSTREAM);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

buf_t *far_bufalloc(uint64_t size, int type, int fd)
{
	buf_t	*result	= (buf_t *)NULL;

	if((type == _FAR_STREAM_BUF) || (type == _FAR_STREAM_FD) ||
		(type == _FAR_STREAM_SD))
	{
		result = calloc(1, sizeof(buf_t));

		if(result)
		{
			switch(type)
			{
				case _FAR_STREAM_BUF:
					result->buffer = calloc(1, size);
					if(result->buffer)
					{
						result->bufsize = size;
					}
					break;

				case _FAR_STREAM_FD:
				case _FAR_STREAM_SD:
					result->fd = fd;
					result->type = type;
					break;

				default:
					if(result)
					{
						if(result->buffer)
						{
							free(result->buffer);
						}
						free(result);
						result = NULL;
					}
			}
		}
	}

	RETURN(result);
}

ferr_t far_read(buf_t *src, void *pdest, uint64_t size, uint64_t *retsize)
{
	register	char_t	*dest = pdest;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dest && src)
	{
		if(size)
		{
			if((far_bufindex(src) + size) > far_bufsize(src))
			{
				size = far_bufsize(src) - far_bufindex(src);
			}

			switch(far_buftype(src))
			{
				case _FAR_STREAM_BUF:
					if(size)
					{
					if(memcpy(dest, far_bufptr(src),
						size))
					{
						far_adv_index(src, size);
						if(retsize)
						{
							(*retsize) = size;
						}
					}
					else
					{
						_far_set_errcode(result,
								errno);
					}
					}
					else
					{
						_far_set_errcode(result,
							_FAR_ERR_SEGFLT);
					}
					break;

				case _FAR_STREAM_FD:
				case _FAR_STREAM_SD:
					if(size > 0)
					result = far_fread(far_filedesc(src),
							pdest, size, retsize);
					break;

				default:
					_far_set_errcode(result,
							_FAR_ERR_BADSTREAM);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t _far_bufseek(buf_t *buf, uint64_t pos)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && (pos >= 0))
	{
		if(_far_isbuffer(buf))
		{
			if(pos < far_bufsize(buf))
			{
				far_set_index(buf, pos);
			}
			else
			{
				_far_set_errcode(result, _FAR_ERR_SEGFLT);
			}
		}
		else if(_far_isfiledesc(buf))
		{
			result = far_lseek(far_filedesc(buf), pos, SEEK_SET);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_init_bufstream(buf_t **stream, char_t *buffer, uint64_t size)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(stream && buffer && (size > 0))
	{
		(*stream) = calloc(1, sizeof(buf_t));

		if((*stream))
		{
			far_setbuf((*stream), buffer, size);
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}
	RETURN(result);
}

ferr_t far_init_fdstream(buf_t **stream, int fd)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(stream && (fd >= 0))
	{
		(*stream) = calloc(1, sizeof(buf_t));

		if((*stream))
		{
			far_setfiledesc((*stream), fd);
			far_settype((*stream), _FAR_STREAM_FD);
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}

	RETURN(result);
}

ferr_t far_init_sdstream(buf_t **stream, int fd)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(stream && (fd >= 0))
	{
		(*stream) = calloc(1, sizeof(buf_t));

		if((*stream))
		{
			far_setfiledesc((*stream), fd);
			far_settype((*stream), _FAR_STREAM_SD);
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}

	RETURN(result);
}

ferr_t far_md5check_buf(buf_t *buf)
{
	char_t	*digest		= (char_t *)NULL;
	char_t	*cdigest	= (char_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf)
	{
		digest = (far_buffer(buf) + (far_bufsize(buf) - 18));
		buf->bufsize-=18; /* || an orrible hack */
		result = far_md5digest(buf, &cdigest);
		buf->bufsize+=18; /* || */

		if(_far_noerror(result))
		{
			if(memcmp(digest, cdigest, 16))
			{
				_far_set_errcode(result, _FAR_ERR_BADCRC);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_md5check_fd(buf_t *buf)
{
	uint64_t	pos		= (uint64_t)0;
	buf_t		*tmp		= (buf_t *)NULL;
	md5_state_t	*md5		= (md5_state_t *)NULL;
	char_t		digest[16];
	char_t		*cdigest	= (char_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf)
	{
		md5 = calloc(1, sizeof(md5_state_t));

		if(md5)
		{
			cdigest = calloc(1, 16);

			if(cdigest)
			{
				pos = _far_bufpos(buf);

				if(pos > 0)
				{
					result = _far_bufseek(buf,
						(far_bufsize(buf) - 18));
				}
				else
				{
					_far_set_errcode(result,
							_FAR_ERR_SEGFLT);
				}

				if(_far_noerror(result))
				{
					tmp = far_bufalloc(_FAR_FBUF_SIZE,
							_FAR_STREAM_BUF, 0);
					result = far_fread(far_filedesc(buf),
							tmp, 16, NULL);

					if(_far_noerror(result))
					{
						if(!memcpy(cdigest,
							far_buffer(buf), 16))
						{
							_far_set_errcode(result, errno);
						}
					}
				}
			}
			else
			{
				_far_set_errcode(result, errno);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}

		if(_far_noerror(result))
		{
			result = _far_bufseek(buf, 0L);

			if(_far_noerror(result))
			{
				result = _far_bufseek(tmp, 0L);
			}

			result = _far_bufseek(buf, 0L);

			if(_far_noerror(result))
			{
				while(_far_noerror(result))
				{
					result = far_fread(far_filedesc(buf),
						tmp, _FAR_FBUF_SIZE, NULL);

					far_md5_append(md5, tmp);
				}

				if(_far_noerror(result) ||
					_far_err_comp(result, _FAR_ERR_EOF))
				{
					md5_finish(md5, digest);
					if(memcmp(cdigest, digest, 16))
					{
						_far_set_errcode(result,
							_FAR_ERR_BADCRC);
					}
				}
			}

			if(cdigest)
			{
				free(cdigest);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

uint64_t _far_bufpos(buf_t *buf)
{
	uint64_t	result = (uint64_t)0;

	if(buf)
	{
		switch(far_buftype(buf))
		{
			case _FAR_STREAM_FD:
				result = lseek(far_filedesc(buf), 0L, SEEK_CUR);
				break;
			case _FAR_STREAM_BUF:
				result = far_bufindex(buf);
				break;
		}
	}

	RETURN(result);
}

